# Python standard library
from functools import partial

# Django Model Signals
from django_model_signals.transceiver import ModelSignalsTransceiver


class ModelSignalsDispatcher:

    @staticmethod
    def signal_method(signal_name, sender, **kwargs):
        if 'instance' in kwargs \
            and isinstance(kwargs['instance'], ModelSignalsTransceiver):
            getattr(kwargs['instance'], signal_name)(**kwargs)
        elif issubclass(sender, ModelSignalsTransceiver):
            getattr(sender, signal_name)(**kwargs)

    @classmethod
    def get_signal_method(cls, name):
        return partial(cls.signal_method, signal_name=name)



# Django
from django.db.models.signals import (
    pre_init,
    post_init,
    pre_save,
    post_save,
    pre_delete,
    post_delete,
    m2m_changed
)
from django.dispatch import Signal


pre_bulk_save = Signal()
post_bulk_save = Signal()
pre_full_clean = Signal()
post_full_clean = Signal()


MODEL_SIGNALS = {
    'pre_init': pre_init,
    'post_init': post_init,
    'pre_full_clean': pre_full_clean,
    'post_full_clean': post_full_clean,
    'pre_save': pre_save,
    'post_save': post_save,
    'pre_delete': pre_delete,
    'post_delete': post_delete,
    'm2m_changed': m2m_changed,
    'pre_bulk_save': pre_bulk_save,
    'post_bulk_save': post_bulk_save
}



# Python Standard Library
import sys

# Django
from django.apps import apps, AppConfig

# Django Model Signals
from django_model_signals.dispatcher import ModelSignalsDispatcher
from django_model_signals.signals import MODEL_SIGNALS
from django_model_signals.transceiver import ModelSignalsTransceiver


class DjangoModelSignalsConfig(AppConfig):
    name = 'django_model_signals'

    def ready(self):

        # Don't bind signals when applying migrations
        if 'manage.py' in sys.argv and 'migrate' in sys.argv:
            return

        models = apps.get_models()
        for model in models:
            if issubclass(model, ModelSignalsTransceiver):
                for signal_name, signal in MODEL_SIGNALS.items():
                    if signal_name in model.ModelSignalsMeta.signals:
                        signal.connect(
                            ModelSignalsDispatcher.get_signal_method(
                                signal_name
                            )
                        )



# Django
from django.db.models.manager import Manager

# Django Model Signals
from django_model_signals.signals import pre_bulk_save, post_bulk_save


class ModelSignalsManager(Manager):

    def bulk_create(
        self,
        objs,
        batch_size=None,
        ignore_conflicts=False,
        update_conflicts=False,
        update_fields=None,
        unique_fields=None
    ):
        pre_bulk_save.send(
            sender=self.get_queryset().model,
            objs=objs,
            created=True,
            batch_size=batch_size,
            ignore_conflicts=ignore_conflicts,
            update_conflicts=update_conflicts,
            update_fields=update_fields,
            unique_fields=unique_fields
        )
        result = super().bulk_create(
            objs,
            batch_size,
            ignore_conflicts,
            update_conflicts,
            update_fields,
            unique_fields
        )
        post_bulk_save.send(
            sender=self.get_queryset().model,
            objs=objs,
            created=True,
            batch_size=batch_size,
            ignore_conflicts=ignore_conflicts,
            update_conflicts=update_conflicts,
            update_fields=update_fields,
            unique_fields=unique_fields,
            result=result
        )
        return result

    def bulk_update(self, objs, fields, batch_size=None):
        pre_bulk_save.send(
            sender=self.get_queryset().model,
            objs=objs,
            fields=fields,
            created=False,
            batch_size=batch_size
        )
        result = super().bulk_update(objs, fields, batch_size)
        post_bulk_save.send(
            sender=self.get_queryset().model,
            objs=objs,
            fields=fields,
            created=False,
            batch_size=batch_size,
            result=result
        )
        return result



from django_model_signals.signals import pre_full_clean, post_full_clean


class ModelSignalsTransceiver:

    @classmethod
    def pre_init(cls, **kwargs):
        pass

    def post_init(self, **kwargs):
        pass

    def pre_full_clean(self, **kwargs):
        pass

    def post_full_clean(self, **kwargs):
        pass

    def post_full_clean_error(self, **kwargs):
        raise kwargs['error']

    def pre_save(self, **kwargs):
        pass

    def post_save(self, **kwargs):
        pass

    def post_save_error(self, **kwargs):
        raise kwargs['error']

    def pre_delete(self, **kwargs):
        pass

    def post_delete(self, **kwargs):
        pass

    def m2m_changed(self, **kwargs):
        pass

    def full_clean(self, *args, **kwargs):
        try:
            signal_kwargs = {
                'instance': self
            }
            if 'pre_full_clean' in self.ModelSignalsMeta.signals:
                pre_full_clean.send(sender=self.__class__, **signal_kwargs)
            super().full_clean(*args, **kwargs)
            if 'post_full_clean' in self.ModelSignalsMeta.signals:
                post_full_clean.send(sender=self.__class__, **signal_kwargs)
            return True
        except Exception as error:
            if 'post_full_clean_error' in self.ModelSignalsMeta.signals:
                return self.post_full_clean_error(
                    error=error,
                    created=self.pk is None
                )
            else:
                raise error

    def save(self, **kwargs):
        try:
            super().save()
        except Exception as error:
            if 'post_save_error' in self.ModelSignalsMeta.signals:
                self.post_save_error(
                    error=error,
                    created=self.pk is None
                )
            else:
                raise error

    @classmethod
    def pre_bulk_save(cls, **kwargs):
      pass

    @classmethod
    def post_bulk_save(cls, **kwargs):
      pass

    class ModelSignalsMeta:
        signals = []
